"""Main module to initiate quke, to compare chat results.

LLMs, embedding model, vector store and other components can be congfigured.
"""
import logging  # functionality managed by Hydra
import os

import hydra
from dotenv import find_dotenv, load_dotenv
from hydra.utils import get_original_cwd, to_absolute_path
from omegaconf import DictConfig, OmegaConf
from rich.console import Console

from . import ClassImportDefinition, ClassRateLimit, DatabaseAction, embed, llm_chat

_ = load_dotenv(find_dotenv())


class ConfigParser:
    """Turns config into a more programming friendly and consistent format.

    For core classes and functions groups configuration by
    class/function arguments for convenience.

    Provides an abstraction layer on top of the structure of the
    Hydra configuration files.
    """

    def __init__(self, cfg: DictConfig) -> None:
        """Creates configuration abstraction."""
        self.cfg = cfg

        # TODO: Read from cfg
        self.src_doc_folder = cfg.source_document_folder

        # TODO: Is this sufficiently robust? What if the user wants a folder not related to wcd/pwd?
        # Consider using os.path.isabs(path) or os.path.abspath(path). Or just pass the string and
        # 'Python will handle it'?
        self.vectordb_location = os.path.join(
            os.getcwd(),
            cfg.internal_data_folder,
            cfg.embedding.vectordb.vectorstore_location,
        )
        self.embedding_import = ClassImportDefinition(
            cfg.embedding.embedding.module_name, cfg.embedding.embedding.class_name
        )
        self.vectordb_import = ClassImportDefinition(
            cfg.embedding.vectordb.module_name, cfg.embedding.vectordb.class_name
        )
        self.llm_import = ClassImportDefinition(
            cfg.llm.module_name_llm, cfg.llm.class_name_llm
        )
        self.embedding_rate_limit = ClassRateLimit(
            cfg.embedding.embedding.rate_limit_chunks,
            cfg.embedding.embedding.rate_limit_delay,
        )
        self.embedding_kwargs = self.get_embedding_kwargs(cfg)

        self.splitter_import = ClassImportDefinition(
            cfg.embedding.splitter.module_name, cfg.embedding.splitter.class_name
        )
        self.splitter_args = self.get_args_dict(cfg.embedding.splitter.args)

        try:
            self.write_mode = DatabaseAction[
                (cfg.embedding.vectordb.vectorstore_write_mode).upper()
            ]
        except Exception:
            logging.warn(
                f"Invalid value configured for cfg.embedding.vectorstore_write_mode: "
                f"{cfg.embedding.vectordb.vectorstore_write_mode}. Using no_overwrite instead."
            )
            self.write_mode = DatabaseAction.NO_OVERWRITE

        self.questions = cfg.question.questions

        try:
            if not cfg.embed_only:
                self.embed_only = False
            else:
                self.embed_only = True
        except Exception:
            self.embed_only = False

        # TODO: need something better for output folder
        # https://hydra.cc/docs/tutorials/basic/running_your_app/working_directory/
        try:  # try statement done for testing suite
            self.output_file = os.path.join(
                hydra.core.hydra_config.HydraConfig.get()["runtime"]["output_dir"],
                cfg.experiment_summary_file,
            )
        except Exception:
            self.output_file = cfg.experiment_summary_file

    def get_embed_params(self):
        """Based on the config files returns the set of parameters need to start embedding."""
        embed_parameters = {
            "src_doc_folder": self.src_doc_folder,
            "vectordb_location": self.vectordb_location,
            "embedding_import": self.embedding_import,
            "embedding_kwargs": self.embedding_kwargs,
            "vectordb_import": self.vectordb_import,
            "rate_limit": self.embedding_rate_limit,
            "splitter_params": self.get_splitter_params(),
            "write_mode": self.write_mode,
        }
        return embed_parameters

    def get_chat_params(self):
        """Based on the config files returns the set of parameters need to start a chat."""
        chat_parameters = {
            "vectordb_location": self.vectordb_location,
            "embedding_import": self.embedding_import,
            "vectordb_import": self.vectordb_import,
            "llm_import": self.llm_import,
            "llm_parameters": self.get_llm_parameters(),
            "prompt_parameters": self.questions,
            "output_file": self.get_chat_session_file_parameters(self.cfg),
        }
        return chat_parameters

    def get_splitter_params(self):
        """Based on the config files returns the set of parameters needed to split source documents."""
        return {
            "splitter_import": self.splitter_import,
            "splitter_args": self.splitter_args,
        }

    def get_args_dict(self, cfg_sub):
        """Takes a subset of the Hydra configs and returns the same as a dict."""
        return OmegaConf.to_container(cfg_sub, resolve=True)

    def get_llm_parameters(self):
        """Based on the config files returns the set of parameters needed to setup an LLM."""
        return OmegaConf.to_container(self.cfg.llm.llm_args, resolve=True)

    def get_chat_session_file_parameters(self, cfg: DictConfig) -> dict:
        """Returns the full configuration in a single yaml and file location for output."""
        chat_sesion_file_parameters = {
            "path": self.output_file,
            "conf_yaml": OmegaConf.to_yaml(cfg),
        }
        return chat_sesion_file_parameters

    def get_embedding_kwargs(self, cfg: DictConfig):
        """Based on the config files returns the set of parameters needed for embedding."""
        try:
            if cfg.embedding.embedding.kwargs:
                embedding_kwargs = cfg.embedding.embedding.kwargs
            else:
                embedding_kwargs = {}
        except Exception:
            embedding_kwargs = {}
        return embedding_kwargs


@hydra.main(version_base=None, config_path="conf", config_name="config")
def slice(cfg: DictConfig) -> None:
    """The main function to initiate a chat.

    Including the embedding of the provided source documents.
    """
    console = Console()
    config_parser = ConfigParser(cfg)

    # ------------------
    print(f"Current working directory : {os.getcwd()}")
    print(f"Orig working directory    : {get_original_cwd()}")
    print(f"to_absolute_path('foo')   : {to_absolute_path('foo')}")
    print(f"to_absolute_path('/foo')  : {to_absolute_path('/foo')}")
    print(f"to_absolute_path('/foo')  : {to_absolute_path(config_parser.output_file)}")
    # return 0
    # ------------------

    embed_parameters = config_parser.get_embed_params()

    with console.status("Embedding...", spinner="aesthetic"):
        # python -m rich.spinner to see options
        embed.embed(**embed_parameters)
        logging.info("\n" + OmegaConf.to_yaml(cfg))

    if not config_parser.embed_only:
        with console.status("Chatting...", spinner="aesthetic"):
            chat_parameters = config_parser.get_chat_params()
            llm_chat.chat(**chat_parameters)


if __name__ == "__main__":
    slice()



"""Compare Large Language Models' capabilities.

Compare the answering capabilities of different LLMs - for example LlaMa, ChatGPT,
Cohere, Falcon - against user provided document(s) and questions.

LLMs, embeddings and other settings can be specified in configuration files.
"""
from collections import namedtuple
from enum import Enum


class DatabaseAction(Enum):
    """Enum for vectorstore write action.

    Value to be specified in config file.

    Action is performed at folder level. For example, if set to OVERWRITE the
    existing folder will be completely deleted before creating a new vector store.
    """

    NO_OVERWRITE = 0
    APPEND = 1
    OVERWRITE = 2


ClassImportDefinition = namedtuple(
    "ClassImportDefinition", ["module_name", "class_name"]
)
ClassRateLimit = namedtuple("ClassRateLimit", ["count_limit", "delay"])



"""Sets up all elements required for a chat session."""
import importlib
import logging  # functionality managed by Hydra
import os
from collections import defaultdict
from datetime import datetime

from langchain.chains import ConversationalRetrievalChain
from langchain.memory import ConversationBufferMemory
from mdutils.fileutils import MarkDownFile
from mdutils.mdutils import MdUtils

from . import ClassImportDefinition


def chat(
    vectordb_location: str,
    embedding_import: ClassImportDefinition,
    vectordb_import: ClassImportDefinition,
    llm_import: ClassImportDefinition,
    llm_parameters,
    prompt_parameters,
    output_file,
) -> object:
    """Initiates a chat with an LLM.

    Sets up all components required for the chat including the LLM,
    embedding model, vector store, chat memory, retriever and actual
    questions.
    """
    module = importlib.import_module(embedding_import.module_name)
    class_ = getattr(module, embedding_import.class_name)
    embedding = class_()

    logging.warn(
        "CAUTION: This function uses external compute services "
        "(like OpenAI or HuggingFace). This is likely to cost money."
    )
    module = importlib.import_module(vectordb_import.module_name)
    class_ = getattr(module, vectordb_import.class_name)
    vectordb = class_(embedding_function=embedding, persist_directory=vectordb_location)

    module = importlib.import_module(llm_import.module_name)
    class_ = getattr(module, llm_import.class_name)
    llm = class_(**llm_parameters)

    memory = ConversationBufferMemory(
        memory_key="chat_history",
        return_messages=True,
        output_key="answer",
    )

    qa = ConversationalRetrievalChain.from_llm(
        llm,
        retriever=vectordb.as_retriever(),
        memory=memory,
        return_source_documents=True,
    )

    # NOTE: trial API keys may have very restrictive rules. It is plausible that you run into
    # constraints after the 2nd question.
    for question in prompt_parameters:
        result = qa({"question": question})

        # print(qa({"question": question}))

        # erik = source_documents
        # print(f"==============={erik}")

        chat_output(question, result)
        chat_output_to_file(result, output_file)

    logging.info("=======================")

    return qa


def chat_output(question: str, result: dict) -> None:
    """Logs a chat question and anwer."""
    logging.info("=======================")
    logging.info(f"Q: {question}")
    logging.info(f"A: {result['answer']}")

    src_docs = [doc.metadata for doc in result["source_documents"]]
    src_docs = dict_crosstab(src_docs, "source", "page")
    for key, value in src_docs.items():
        logging.info(f"Source document: {key}, Pages used: {value}")


# TODO: Either I do not understand mdutils or it is an unfriendly package when trying to append.
def chat_output_to_file(result: dict, output_file) -> None:
    """Populates a record of the chat with the LLM into a markdown file."""
    first_write = not os.path.isfile(output_file["path"])

    mdFile = MdUtils(file_name="tmp.md")

    if first_write:
        mdFile.new_header(1, "LLM Chat Session with quke")
        mdFile.write(
            datetime.now().astimezone().strftime("%a %d-%b-%Y %H:%M %Z"), align="center"
        )
        mdFile.new_paragraph("")
        mdFile.new_header(2, "Experiment settings", header_id="settings")
        mdFile.insert_code(output_file["conf_yaml"], language="yaml")
        mdFile.new_header(2, "Chat", header_id="chat")
    else:
        existing_text = MarkDownFile().read_file(file_name=output_file["path"])
        mdFile.new_paragraph(existing_text)

    mdFile.new_paragraph(f"Q: {result['question']}")
    mdFile.new_paragraph(f"A: {result['answer']}")

    src_docs = [doc.metadata for doc in result["source_documents"]]
    src_docs = dict_crosstab(src_docs, "source", "page")
    for key, value in src_docs.items():
        mdFile.new_paragraph(f"Source document: {key}, Pages used: {value}")

    new = MarkDownFile(name=output_file["path"])

    new.append_end((mdFile.get_md_text()).strip())


def dict_crosstab(source, key, listed, missing="NA"):
    """limited and simple version of a crosstab query on a dict."""
    dict_subs = []
    for d in source:
        dict_subs.append({key: d[key], listed: d.get(listed, missing)}.values())

    d = defaultdict(list)
    for k, v in dict_subs:
        d[k].append(v)
        d[k] = list(set(d[k]))

    return dict(d)  # TODO: consider sorted(d.items())
